const config = require("config");
const { pki, random, md } = require("node-forge");
const crypto = require("crypto");
const BlindSignature = require("blind-signatures");
var BigInteger = require("jsbn").BigInteger;
const NodeRSA = require("node-rsa");

class Crypto {
  #dayjs;
  #dayjs_format;
  #redis;

  constructor() {
    this.#dayjs = require("dayjs");
    this.#dayjs_format = config.get("crypto.private_key.date_format");
    let Redis = require("ioredis");
    this.#redis = new Redis({
      host: config.get("redis.host"),
    });
  }

  async private_key_get(seed_date, expiration_date) {
    let cache_key =
      "private_key_" +
      seed_date.format(this.#dayjs_format) +
      "_" +
      expiration_date.format(this.#dayjs_format);
    // Check if the private key already exists in cache
    let cached_key = await this.#redis.get(cache_key);
    let private_key = new NodeRSA();
    if (cached_key === null) {
      let seed =
        config.get("crypto.private_key.seed") +
        seed_date.format(this.#dayjs_format) +
        "_" +
        expiration_date.format(this.#dayjs_format);
      let sha512 = md.sha512.create();
      sha512.update(seed);
      seed = sha512.digest().toHex();
      // Feed the seed into a pseudorandom number generator
      let prng_instance = random.createInstance();
      prng_instance.seedFileSync = () => seed;

      let keypair = pki.rsa.generateKeyPair({
        bits: config.get("crypto.private_key.bit_length"),
        prng: prng_instance,
        workers: -1,
      });
      let private_key_pem = pki
        .privateKeyToPem(keypair.privateKey)
        .trim()
        .replace(/\r\n/g, "\n");

      private_key = private_key.importKey(
        Buffer.from(private_key_pem, "utf8"),
        "private"
      );

      // Store the key in cache
      await this.#redis.set(cache_key, private_key.exportKey("pkcs8"));
    } else {
      let private_key_data = await this.#redis.get(cache_key);
      private_key = private_key.importKey(private_key_data, "pkcs8");
    }
    return private_key;
  }

  async validateMetaGerPassCode(
    generation_month,
    expiration_month,
    metager_pass_codes
  ) {
    // Check if codes are expired
    if (!this.#dayjs().isBefore(expiration_month, "month")) {
      return Promise.reject("Redeem Codes are expired.");
    }

    let private_key = await this.private_key_get(
      generation_month,
      expiration_month
    );

    let uuid_regexExp =
      /^[0-9a-fA-F]{8}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{12}$/gi;
    for (let i = 0; i < metager_pass_codes.length; i++) {
      // Check if code iss in correct format
      let metager_pass_code = metager_pass_codes[i];
      if (
        !metager_pass_code.hasOwnProperty("code") ||
        !metager_pass_code.code.match(uuid_regexExp)
      ) {
        return Promise.reject(metager_pass_code.code + " is not a valid UUID");
      }
      // Check if signature is in correct format
      if (
        !metager_pass_code.hasOwnProperty("signature") ||
        !metager_pass_code.signature.match(/^\d+$/gi)
      ) {
        return Promise.reject(i + 1 + ". signature has not a valid format");
      }

      let verification_result = BlindSignature.verify2({
        unblinded: metager_pass_code.signature,
        key: private_key,
        message: metager_pass_code.code,
      });
      if (!verification_result) {
        console.log("Failed" + i);
        return Promise.reject("One or more signatures could not be verified");
      }
    }
  }

  async sign(encrypted_sales_receipts, order_month, expiration_month) {
    let private_key = await this.private_key_get(order_month, expiration_month);
    let signed_encrypted_sales_receipts = [];

    for (let i = 0; i < encrypted_sales_receipts.length; i++) {
      signed_encrypted_sales_receipts.push(
        BlindSignature.sign({
          blinded: new BigInteger(encrypted_sales_receipts[i]),
          key: private_key,
        }).toString()
      );
    }
    return signed_encrypted_sales_receipts;
  }

  /**
   * Creates an hmac hash for purchase data so we can check it later
   */
  createIntegrityHash(
    order_id,
    expires_at,
    amount,
    unit_size,
    price_per_unit,
    public_key_n,
    public_key_e
  ) {
    let data_to_hash = JSON.stringify({
      order_id: parseInt(order_id),
      expires_at: new String(expires_at),
      amount: parseInt(amount),
      unit_size: parseInt(unit_size),
      price_per_unit: parseFloat(price_per_unit),
      public_key_n: new String(public_key_n),
      public_key_e: new String(public_key_e),
    });
    let forge = require("node-forge");
    let hmac = forge.hmac.create();
    hmac.start("sha256", config.get("crypto.hmac_integrity_seed"));
    hmac.update(data_to_hash);
    return hmac.digest().toHex();
  }

  /**
   * Validates an hmac hash for purchase data so we can check it later
   */
  validateIntegrityHash(
    user_hash,
    order_id,
    expires_at,
    amount,
    unit_size,
    price_per_unit,
    public_key_n,
    public_key_e
  ) {
    let data_to_hash = JSON.stringify({
      order_id: parseInt(order_id),
      expires_at: new String(expires_at),
      amount: parseInt(amount),
      unit_size: parseInt(unit_size),
      price_per_unit: parseFloat(price_per_unit),
      public_key_n: new String(public_key_n),
      public_key_e: new String(public_key_e),
    });
    let forge = require("node-forge");
    let hmac = forge.hmac.create();
    hmac.start("sha256", config.get("crypto.hmac_integrity_seed"));
    hmac.update(data_to_hash);
    let server_hash = hmac.digest().toHex();

    return crypto.timingSafeEqual(
      Buffer.from(server_hash, "utf8"),
      Buffer.from(user_hash, "utf8")
    );
  }
}

module.exports = Crypto;
