const dayjs = require("dayjs");
const RedisClient = require("./RedisClient");

class Key {
  static get DATABASE_PREFIX() {
    return "key:";
  }

  static get MAX_ORDERS_PER_KEY() {
    return 3;
  }

  static get EXPIRATION_AFTER_CHARGE_DAYS() {
    return 365;
  }

  /**
   * @type {String}
   */
  #key;

  /**
   * @type {Boolean}
   */
  #writable;
  /**
   * @typedef {Object} KeyOrder
   * @property {Number} amount
   * @property {dayjs.Dayjs} expiration
   * @property {Number} order_id
   */
  /**
   * @type {Array<KeyOrder>}
   */
  #key_orders = [];

  /**
   * 
   * @param {String} key 
   * @param {Array<KeyOrder>} key_data 
   * @param {Boolean} writable 
   */
  constructor(key, key_data, writable = false) {
    this.#key = key;
    this.#writable = writable;
    this.#key_orders = key_data;
  }

  /**
   * Returns the string representation of this key
   * 
   * @returns {String}
   */
  get_key() {
    return this.#key;
  }

  /**
   * Returns the sum of all stored key orders
   * 
   * @param {String} order_id - Optional
   * 
   * @returns {Number}
   */
  get_charge(order_id = null) {
    let sum = 0;
    this.#key_orders.forEach(key_order => {
      if (order_id === null || order_id === key_order.order_id) {
        sum += key_order.amount;
      }
    });
    return sum;
  }

  /**
   * Returns the date when all order keys will have been expired
   * 
   * @returns {dayjs.Dayjs}
   */
  get_expiration_date() {
    /**
     * @type {dayjs.Dayjs}
     */
    let expiration = null;
    this.#key_orders.forEach(key_order => {
      if (expiration === null) {
        expiration = key_order.expiration;
      } else if (expiration.isBefore(key_order.expiration)) {
        expiration = key_order.expiration;
      }
    });
    if (expiration === null) {
      expiration = dayjs().add(Key.EXPIRATION_AFTER_CHARGE_DAYS, "day");
    }
    return expiration;
  }

  /**
   * 
   * @returns {KeyOrder}
   */
  get_charge_orders() {
    return this.#key_orders;
  }

  /**
   * Charges this key in response to an order
   * 
   * @param {Number} amount 
   * @param {Number} order_id 
   */
  charge_key_order(amount, order_id) {
    let order_charged = false;
    // If we recharge an existing order
    this.#key_orders.forEach(key_order => {
      if (key_order.order_id === order_id) {
        key_order.amount += amount;
        order_charged = true;
      }
    });
    // Otherwise create new Order
    if (!order_charged) {
      this.#key_orders.push({
        amount: amount,
        order_id: order_id,
        expiration: dayjs().add(Key.EXPIRATION_AFTER_CHARGE_DAYS, "day")
      });
    }
  }

  /**
   * Discharges this key for a specific order
   * 
   * @param {Number} amount 
   * @param {Number} order_id 
   * 
   * @returns
   */
  discharge_key(amount, order_id = null) {
    let discharged_amount = null;
    this.#key_orders.every((key_order, index) => {
      if (order_id === null || key_order.order_id === order_id) {
        if (amount >= key_order.amount) {
          // Key cannot be reduced by more than its current charge
          discharged_amount = key_order.amount;
        } else {
          discharged_amount = amount;
        }
        this.#key_orders[index].amount -= discharged_amount;
        return false;
      } else {
        return true;
      }
    });
    if (discharged_amount !== null) {
      return discharged_amount;
    } else {
      throw new Error(`Cannot find order ${order_id} for discharge`);
    }
  }

  async save() {
    if (!this.#writable) {
      throw new Error("Cannot write to readonly Key");
    }
    let write_lock_key = Key.DATABASE_PREFIX + this.#key + ":writelock";
    return RedisClient.CLIENT()
      .pipeline()
      .setex(Key.DATABASE_PREFIX + this.#key, this.get_expiration_date().diff(dayjs(), "second"), JSON.stringify(this.#key_orders))
      .del(write_lock_key)
      .exec().then(() => this);
  }

  /**
   * Resolves with a Key object
   * 
   * @param {string} key 
   * @param {boolean} writable
   * 
   * @return {Promise<Key>}
   */
  static async GET_KEY(key, writable = false) {
    // Check if supplied key is UUID v4
    if (!key.match(/^[0-9A-F]{8}-[0-9A-F]{4}-[4][0-9A-F]{3}-[89AB][0-9A-F]{3}-[0-9A-F]{12}$/i)) {
      const { createHash } = await import("node:crypto");
      let hash = createHash("md5");
      hash.update(key);

      // According to the definition of UUIDv4 the 13th digit needs to specify the uuid version (i.e. 4 in our case)
      // Additionally the 17th digit needs to specify its variant (DCE 1.1, ISO/IEC 11578:1996 in our case)
      // This means the binary representation of the 17th digit needs to be 10xx with the x's being random
      // This means we need a decimal number between 8 and 11 
      let variant = 0
      for (let i = 0; i < key.length; i++) {
        variant += key.charCodeAt(i);
      }
      variant = variant % 4;
      // Convert decimal to hex
      variant = 8 + variant;
      variant = variant.toString(16);
      key = hash.copy().digest('hex').replace(/^([a-f0-9]{8})([a-f0-9]{4})[a-f0-9]{1}([a-f0-9]{3})[a-f0-9]{1}([a-f0-9]{3})([a-f0-9]{12})/, "$1-$2-4$3-" + variant + "$4-$5");
    }
    return RedisClient.CLIENT().get(Key.DATABASE_PREFIX + key).then(key_data => {
      if (!key_data) {
        return [];
      }
      key_data = JSON.parse(key_data);
      key_data.forEach((key_order, index) => {
        key_data[index].expiration = dayjs(key_order.expiration);
      })
      return key_data
    }).then(async (key_data) => {
      if (writable) {
        // Acquire an exclusive lock for this key entry
        await Key.GET_WRITELOCK(key);
      }
      return new Key(key, key_data, writable);
    });
  }

  static async GET_WRITELOCK(key) {
    let write_lock_key = Key.DATABASE_PREFIX + key + ":writelock";

    return new Promise(async (resolve, reject) => {
      let timestart = dayjs();
      let write_lock = 0;
      do {
        write_lock = await RedisClient.CLIENT()
          .pipeline()
          .setnx(write_lock_key, 1)
          .expiretime(write_lock_key)
          .expire(write_lock_key, 15)
          .exec();
        let expire_at = write_lock[1][1];
        write_lock = write_lock[0][1];
        if (write_lock === 1) {
          resolve(true);
          break;
        } else if (dayjs().diff(timestart, "second") >= 15) {
          if (expire_at >= 0) {
            await RedisClient.CLIENT().expireat(write_lock_key, expire_at);
          }
          reject("Timed out waiting for write lock for Key");
          break;
        } else {
          if (expire_at >= 0) {
            await RedisClient.CLIENT().expireat(write_lock_key, expire_at);
          }
          await new Promise((resolve) => setTimeout(resolve, 1000));
        }
      } while (true);
    });
  }

  static async EXISTS(key) {
    return new Promise((resolve, reject) => {
      let redis_client = RedisClient.CLIENT();
      redis_client.exists(Key.DATABASE_PREFIX + key).then((result) => {
        resolve(result);
      });
    });
  }

  /**
   *
   * @returns A new MetaGer-Pass Key with 0 searches
   */
  static async GET_NEW_KEY() {
    return new Promise(async (resolve, reject) => {
      let redis_client = RedisClient.CLIENT();
      let crypto = require("crypto");
      do {
        let key = crypto.randomUUID();

        let key_exists = await redis_client.exists(Key.DATABASE_PREFIX + key);
        if (!key_exists) {
          return resolve(key);
        }
        await new Promise((resolve) => setTimeout(resolve, 50));
      } while (true);
    });
  }

  static async CHARGE_KEY_ORDER(key, amount, order_id) {
    let redis_client = RedisClient.CLIENT();
    let expiration = require("dayjs")().add(
      Key.EXPIRATION_AFTER_CHARGE_DAYS,
      "day"
    );
    // Check if key exists and is eligable for recharge
    return redis_client
      .pipeline()
      .incrby(Key.DATABASE_PREFIX + key, amount)
      .expireat(Key.DATABASE_PREFIX + key, expiration.unix())
      .exec();
  }

  static async DISCHARGE_KEY(key, amount) {
    let redis_client = RedisClient.CLIENT();

    // Dicharge key and make sure it can't fall below 0
    return redis_client
      .decrby(Key.DATABASE_PREFIX + key, amount)
      .then((new_charge) => {
        if (new_charge < 0) {
          return redis_client
            .incrby(Key.DATABASE_PREFIX + key, Math.abs(new_charge))
            .then(() => 0);
        } else {
          return new_charge;
        }
      });
  }
}

module.exports = Key;
