var express = require("express");
var router = express.Router();
const config = require("config");
const { auth, requiresAuth, claimCheck } = require("express-openid-connect");

const auth_session_expiration_seconds = 600;

/**
 * Authorization is done using Oauth2 via any Openid Provider
 * The Instance to be used can be configured in the config.
 *
 * You can call any url with the parameter moderation=true to trigger a login with
 * redirect to the same URL after successfull authentication or 401 error on
 * failed authentication.
 */
router.use(
  // General Optional authorization for every page
  // It will not require login or start a session by default
  // Only when a route uses the middleware requireAuth
  (req, res, next) => {
    auth({
      issuerBaseURL: `${config.get("app.openid_auth.url")}`,
      baseURL: res.locals.baseDir,
      clientID: config.get("app.openid_auth.app_id"),
      clientSecret: config.get("app.openid_auth.app_secret"),
      secret: config.get("app.secret"),
      session: {
        rollingDuration: auth_session_expiration_seconds,
        signSessionStoreCookie: true,
      },
      authRequired: false,
    })(req, res, next);
  },
  /**
   * If the user is already authenticated this middleware will check
   * if the user inherits the required role in his oidc token to use this app
   */
  (req, res, next) => {
    if (req.data === undefined) {
      req.data = { admin: false };
    } else {
      req.data.admin = false;
    }
    if (req.oidc.isAuthenticated()) {
      claimCheck((req, claims) => {
        req.data.admin = true;
        return true;
      })(req, res, next);
    } else {
      next();
    }
  },
  // Require authorization for every route with moderation request parameter set
  // Or when the user already is authenticated
  (req, res, next) => {
    if (req.query.moderation || req.data.admin) {
      requiresAuth()(req, res, next);
    } else {
      next("route");
    }
  },
  // Redirect the user to the same URL without moderation URL Parameter after
  // successful auth. Show him a 401 if he does not inherit required role
  (req, res, next) => {
    if (req.query.moderation) {
      if (!req.data.admin) {
        res.locals.error = { status: 401 };
        res.locals.message = "Unauthorized";
        res.status(401).render("error");
      } else {
        let params = req.query;
        delete params.moderation;
        let redirect_url = `${res.locals.baseDir}${req.path}?${new URLSearchParams(
          params
        ).toString()}`;
        res.redirect(redirect_url);
      }
    } else {
      next("route");
    }
  }
);

module.exports = router;
