var express = require("express");
var router = express.Router({ mergeParams: true });

const config = require("config");
const Order = require("../../app/Order.js");
const Key = require("../../app/Key.js");
const Paypal = require("../../app/payment_processor/Paypal.js");

router.use("/", async (req, res, next) => {
  if (req.data && req.data.checkout) {
    req.data.checkout.payment = {
      provider: "paypal",
      paypal: {
        client_id: config.get(`payments.paypal.client_id`),
      },
    };
  }
  new Paypal()
    .verifyWebhook()
    .then(() => {
      next("route");
    })
    .catch((reason) => {
      console.error(reason);
      next("route");
    });
});

router.get("/:funding_source", async (req, res) => {
  req.data.checkout.payment.paypal.funding_source = req.params.funding_source;

  res.header({ "Access-Control-Allow-Origin": "https://www.paypal.com" });
  if (req.params.funding_source === "card") {
    let client_token = await new Paypal().generateClientToken();
    req.data.checkout.payment.paypal.client_token = client_token;
  }

  req.data.change_url.funding_source =
    `${res.locals.baseDir}/key/` +
    encodeURIComponent(req.data.key.key.get_key()) +
    "/checkout/" +
    encodeURIComponent(req.data.checkout.amount) +
    "#payment";

  req.data.change_url.funding_source_not_eligible =
    `${res.locals.baseDir}/key/` +
    encodeURIComponent(req.data.key.key.get_key()) +
    "/checkout/" +
    encodeURIComponent(req.data.checkout.amount) +
    "?error=funding_source_not_eligible";

  req.data.change_url.order_base_url =
    `${res.locals.baseDir}/key/` +
    encodeURIComponent(req.data.key.key.get_key()) +
    "/checkout/" +
    encodeURIComponent(req.data.checkout.amount) +
    "/paypal/" +
    encodeURIComponent(req.data.checkout.payment.paypal.funding_source) +
    "/order";

  req.data.js.push(`${res.locals.baseDir}/js/checkout_paypal.js`);
  res.render("key", req.data);
});

router.post("/:funding_source/order/create", async (req, res) => {
  // Order data is validated: Create and store the order in the redis database
  Order.CREATE_NEW_ORDER(
    // Create Order on our side
    req.params.amount,
    (req.params.amount / 300) * config.get("price.per_300"),
    req.data.key.key.get_key(),
    new Paypal()
  )
    .then(
      /** @param {Order} */(order) => {
        res.status(200).json({
          order_id: order.getOrderID(),
          paypal_order_id: order.getPaymentProcessor().getOrderId(),
        });
      }
    )
    .catch((reason) => {
      console.error(reason);
      res.status(400).json({
        errors: [{ msg: "Failed to create a new Order. Try again later" }],
      });
    });
});

router.post("/:funding_source/order/cancel", async (req, res) => {
  Order.LOAD_ORDER_FROM_ID(req.body.order_id)
    .then(
      /** @param {Order} order */(order) => {
        if (order.isPaymentCaptured()) {
          // Not so good. Something went wrong after we captured the Payment
          // Refund it back
          // ToDo Execute refund
          console.debug("Refund should happen here");
        }
        return order.delete();
      }
    )
    .then((success) => {
      if (success) {
        res.status(200).json({ msg: "Order deleted" });
      } else {
        res.status(400).json({ msg: "Order could not be deleted" });
      }
    })
    .catch((reason) => {
      console.error(reason);
      res.status(400).json({ msg: "Failed to load/cancel Order." });
    }); // Deletes a order but only if the payment is not yet completed
});

// We need to make sure that a card payment was authorized with 3-DS.
// All other PayPal captures will be handled in the next capture rotue
router.use("/card/order/capture", async (req, res, next) => {
  Order.LOAD_ORDER_FROM_ID(req.body.order_id)
    .then((order) => order.getPaymentProcessor().verify_3D())
    .then(() => next("route"))
    .catch((reason) => {
      console.debug(reason);
      res.status(400).json({
        errors: [
          { type: "PAYPAL_CARD_3D_ERROR", msg: "3D Verification Failed" },
        ],
      });
    });
});

// capture payment & store order information or fullfill order
router.post("/:funding_source/order/capture", async (req, res) => {
  Order.LOAD_ORDER_FROM_ID(req.body.order_id).then(
    /**
     * @param {Order} loaded_order
     */
    (loaded_order) => {
      loaded_order
        .captureOrder()
        .then(() => loaded_order.chargeKey())
        .then(() => {
          let redirect_url =
            `${res.locals.baseDir}/key/` + req.data.key.key.get_key() + "/orders/" + loaded_order.getOrderID();
          res.status(200).json({
            redirect_url: redirect_url,
            order: {
              id: loaded_order.getOrderID(),
              amount: loaded_order.getAmount(),
              price: loaded_order.getPrice(),
            },
          });
        })
        .catch((reason) => {
          console.debug(reason);
          res.status(400).json({
            errors: [
              {
                type: "PAYMENT_NOT_COMPLETED_ERROR",
                msg: "Couldn't capture the payment",
              },
            ],
          });
        });
    }
  );
});

router.post("/webhook", async (req, res) => {
  // Verify that the webhook came from paypal
  let paypal = new Paypal();
  paypal
    .processWebhook(
      req.headers["paypal-auth-algo"],
      req.headers["paypal-cert-url"],
      req.headers["paypal-transmission-id"],
      req.headers["paypal-transmission-sig"],
      req.headers["paypal-transmission-time"],
      req.body
    )
    .then(() => {
      res.status(200).send("");
    })
    .catch((reason) => {
      console.debug(reason);
      res.status(200).send("");
    });
});

module.exports = router;

//////////////////////

// PayPal API helpers

//////////////////////

async function refundPayment(capture_ids) {
  const accessToken = await generateAccessToken();
  let promises = [];
  capture_ids.forEach((capture_id) => {
    promises.push(
      fetch(`${base}/v2/payments/captures/${capture_id}/refund`, {
        method: "post",
        headers: {
          Authorization: `Bearer ${accessToken}`,
          "Content-Type": "application/json",
        },
        body: JSON.stringify({
          note_to_payer: "Something went wrong when processing your payment.",
        }),
      })
    );
  });
  return Promise.all(promises);
}

// generate an access token using client id and app secret
