var express = require("express");
var router = express.Router({ mergeParams: true });

const config = require("config");
const { body, validationResult } = require("express-validator");
const OrderReceipt = require("../../app/pdf/OrderReceipt");

router.get("/", (req, res) => {
  req.data.order.invoice = {
    params: {
      name: req.query.name || "",
      email: req.query.email || "",
      address: req.query.address || "",
    },
    create_invoice_url: `${res.locals.baseDir}/key/${req.data.key.key.get_key()
      }/orders/${req.data.order.order.getOrderID()}/invoice/create`,
    errors: {},
  };

  res.render("key", req.data);
});

router.post(
  "/*",
  body("email").isEmail({ domain_specific_validation: true }),
  body("address").isLength({ max: 1000 }),
  body("name").isLength({ max: 500 }),
  (req, res, next) => {
    req.data.order.invoice = {
      params: {
        name: req.body.name || "",
        email: req.body.email || "",
        address: req.body["address"] || "",
      },
      errors: {},
    };

    const errors = validationResult(req);
    if (!errors.isEmpty()) {
      errors.array().forEach((error) => {
        req.data.order.invoice.errors[error.param] = error.msg;
      });
      res.render("key", req.data);
      return;
    }
    next();
  }
);

router.post(
  "/",
  (req, res, next) => {
    // Check if admin parameter is set and if so if the user is authenticated
    if (req.body.admin) {
      if (req.data.admin) {
        next();
      } else {
        res.locals.error = { status: 401 };
        res.locals.message = "Unauthorized";
        res.status(401).render("error");
      }
    } else {
      next();
    }
  },
  (req, res) => {
    let moderation_params = {
      moderation: true,
      name: req.data.order.invoice.params.name,
      email: req.data.order.invoice.params.email,
      address: req.data.order.invoice.params.address,
    };

    req.data.order.invoice.moderation_url = `${res.locals.baseDir}/key/${req.data.key.key.get_key()
      }/orders/${req.data.order.order.getOrderID()}/invoice?${new URLSearchParams(
        moderation_params
      ).toString()}#invoice-form`;

    // Render the message
    let ejs = require("ejs"),
      fs = require("fs"),
      template = fs.readFileSync(
        `${__dirname}/../../views/orders/invoice_message.ejs`,
        "utf-8"
      );

    let message = ejs.render(template, req.data);

    // No validation errors. Try to create a new Ticket
    return fetch(`${config.get("app.osticket.url")}/api/tickets.json`, {
      method: "post",
      headers: {
        "X-API-Key": config.get("app.osticket.api_key"),
        "Content-Type": "application/json",
      },
      body: JSON.stringify({
        alert: true,
        autorespond: true,
        source: "API",
        name: req.data.order.invoice.params.name,
        email: req.data.order.invoice.params.email,
        subject: `MetaGer Schlüssel: Rechnung (${req.data.order.order.getOrderID()})`,
        message: `data:text/html;charset=utf-8,${message}`,
        topicId: 12, // ToDo change topic for english autoresponder
      }),
    })
      .then((response) => {
        if (response.status != 201) {
          console.error(response.body);
          throw "Fehler beim Erstellen der Benachrichtigung,";
        } else {
          req.data.order.invoice.success = true;
          res.render("key", req.data);
        }
      })
      .catch((reason) => {
        console.log(reason);
        req.data.order.invoice.errors["send_email"] =
          "Fehler beim Erstellen der Benachrichtigung,";
        res.render("key", req.data);
      });
  }
);

router.post(
  "/create",
  (req, res, next) => {
    // Check if admin parameter is set and if so if the user is authenticated
    if (req.data.admin) {
      next();
    } else {
      res.locals.error = { status: 401 };
      res.locals.message = "Unauthorized";
      res.status(401).render("error");
    }
  },
  (req, res, next) => {
    if (req.data.order.order.isReceiptCreated()) {
      res.locals.error = { status: 400 };
      res.locals.message = "Invoice already created";
      res.status(400).render("error");
    } else {
      next();
    }
  },
  (req, res) => {
    OrderReceipt.CREATE_ORDER_RECEIPT(
      req.data.order.order,
      req.data.order.invoice.params
    )
      .then((data) => {
        let b64_string = Buffer.concat(data).toString("base64");
        return req.data.order.order.attachReceipt(b64_string);
      })
      .then(() => {
        // Added receipt. We can redirect to download page
        res.redirect(
          `${res.locals.baseDir}/key/${req.data.key.key.get_key()
          }/orders/${req.data.order.order.getOrderID()}/invoice/download`
        );
      })
      .catch((reason) => {
        res.locals.error = { status: 500, stack: reason };
        res.locals.message = "Failed to create invoice in Civicrm";
        res.status(500).render("error");
      });
  }
);

router.get("/download", (req, res) => {
  if (req.data.order.order.isReceiptCreated()) {
    req.data.order.order
      .getReceipt()
      .then((receipt) => {
        receipt = receipt.values[0];
        res
          .header({
            "Content-Type": receipt.file_mime_type,
            "Content-Disposition": `attachment; filename=INV_${req.data.order.order.getOrderID()}.pdf`,
          })
          .send(Buffer.from(receipt.file_content, "base64"));
      })
      .catch((reason) => {
        console.log(reason);
        let status = 400;
        if (reason.error_code) {
          status = reason.error_code;
        }
        res.locals.error = {
          status: status,
          stack: JSON.stringify(reason, null, 4),
        };
        res.locals.message = "Error fetching receipt";
        res.status(status).render("error");
      });
  } else {
    res.locals.error = { status: 404 };
    res.locals.message = "Invoice not found";
    res.status(404).render("error");
  }
});

module.exports = router;
