var express = require("express");
var router = express.Router();
const { query, body, validationResult } = require("express-validator");

const config = require("config");
const dayjs = require("dayjs");
const Crypto = require("../app/Crypto");
var customParseFormat = require("dayjs/plugin/customParseFormat");
const Key = require("../app/Key");
const Order = require("../app/Order");
const path = require('path');
const fs = require('fs');
const readline = require('readline');
dayjs.extend(customParseFormat);

/**
 * This routes are called after a purchase and intended to
 * exchange a purchase receipt against a MetaGer-Pass Key
 *
 * This middleware validates the purchase receipt for all of the other routes
 */
router.use(
  "/",
  body("expiration_month")
    .matches(/^\d{4}-\d{2}$/)
    .custom((expiration_month, { req }) => {
      dayjs(expiration_month + "-01");
      return true;
    })
    .customSanitizer(expiration_month => {
      return dayjs(expiration_month);
    })
    .withMessage("Invalid Expiration Month supplied"),
  body("generation_month")
    .matches(/^\d{4}-\d{2}$/)
    .custom((generation_month, { req }) => {
      return dayjs(generation_month + "-01");
    })
    .customSanitizer(generation_month => {
      return dayjs(generation_month);
    })
    .withMessage("Invalid Generation Month supplied"),
  body("receipts").custom(async (receipts, { req }) => {
    return new Promise((resolve, reject) => {
      new Crypto()
        .validateMetaGerPassCode(
          req.body.generation_month,
          req.body.expiration_month,
          req.body.metager_pass_codes
        )
        .then(() => {
          resolve(true);
        })
        .catch((reason) => {
          reject(reason);
        });
    });
  }),
  (req, res, next) => {
    const errors = validationResult(req);
    if (!errors.isEmpty()) {
      return res.status(400).json({ errors: errors.array() });
    }
    next("route");
  }
);

/* Recharge a MetaGer-Pass Key */
router.post("/", body("metager_pass_key").isWhitelisted(Key.KEY_CHARSET).isLength({ min: 6, max: 20 }), async (req, res, next) => {
  const errors = validationResult(req);
  if (!errors.isEmpty()) {
    return res.status(400).json({ errors: errors.array() });
  }

  let max_recharge_tries = 10;
  let max_recharge_tries_hours = 1;   // Defines in which time frame the recharge tries are counted; recharge tries get reset after this amount of hours without recharge try have passed

  let charge_amount = 0;

  // Redis Client
  let Redis = require("ioredis");
  let redis_client = new Redis({
    host: config.get("redis.host"),
  });
  // Dayjs
  let dayjs = require("dayjs");

  let key_recharge_cache_keys = [];
  let key_recharge_cache_prefix = "recharge_key_cache";
  for (let i = 0; i < req.body.metager_pass_codes.length; i++) {
    let code = req.body.metager_pass_codes[i].code;
    key_recharge_cache_keys.push(key_recharge_cache_prefix + "_" + code);
    charge_amount += Order.PACKET_SIZE;
  }

  // Check one or more of the codes was already used to redeem a MetaGer-Pass Key
  let order_month = dayjs(req.body.generation_month);
  let redeem_file_path = path.join(
    config.get("storage.data_path"),
    process.env.NODE_ENV,
    order_month.format("YYYY"),
    order_month.format("MM"),
    "redeemed.json");
  if (!fs.existsSync(path.dirname(redeem_file_path))) {
    fs.mkdirSync(path.dirname(redeem_file_path), { recursive: true });
  }

  if (fs.existsSync(redeem_file_path)) {
    let rl = readline.createInterface({
      input: fs.createReadStream(redeem_file_path),
      output: process.stdout,
      terminal: false
    });
    for await (const line of rl) {
      for (let i = 0; i < req.body.metager_pass_codes.length; i++) {
        if (req.body.metager_pass_codes[i].code === line.trim()) {
          return res.status(400).json({
            status: "FAILED",
            msg: ["One or more of the provided MetaGer-Pass Codes are already redeemed."]
          });
        }
      }
    }
  }


  let recharge_tries = 0;
  redis_client.mget(key_recharge_cache_keys).then(async response => {
    for (let i = 0; i < response.length; i++) {
      let tries = response[i];
      if (tries && tries > recharge_tries) {
        recharge_tries = tries;
      }
    }

    if (recharge_tries >= max_recharge_tries) {
      res.status(400).json({
        status: "FAILED",
        msg: "Too many failed attempts against non existing MetaGer-Pass Keys. Please try again later."
      });
    } else {
      recharge_tries++;

      let redis_pipeline = redis_client.pipeline();
      let expiration = dayjs().add(max_recharge_tries_hours, "hour");
      for (let i = 0; i < key_recharge_cache_keys.length; i++) {
        redis_pipeline.set(key_recharge_cache_keys[i], recharge_tries);
        redis_pipeline.expireat(key_recharge_cache_keys[i], expiration.unix());
      }
      await redis_pipeline.exec();

      Key.CHARGE_EXISTING_KEY(req.body.metager_pass_key, charge_amount).then(result => {
        // Key is charged store the redeem codes into filesystem so they can only be used once
        for (let i = 0; i < req.body.metager_pass_codes.length; i++) {
          fs.appendFileSync(redeem_file_path, req.body.metager_pass_codes[i].code + "\n");
        }

        res.json(result);
      }).catch(reason => {
        res.status(400).json({
          status: "FAILED",
          msg: [reason]
        });
      });
    }
  });
});

/* Create a new MetaGer-Pass Key from purchase receipt */
router.post("/create", async (req, res, next) => {
  let key = "";

  // Redis Client
  let Redis = require("ioredis");
  let redis_client = new Redis({
    host: config.get("redis.host"),
  });
  // Dayjs
  let dayjs = require("dayjs");

  let key_cache_redis_keys = [];
  let key_redis_cache_prefix = "create_key_cache";
  // Check if the user already created a key which is in redis
  for (let i = 0; i < req.body.metager_pass_codes.length; i++) {
    let code = req.body.metager_pass_codes[i].code;
    key_cache_redis_keys.push(key_redis_cache_prefix + "_" + code);
  }
  redis_client.mget(key_cache_redis_keys).then(async response => {
    let existing_key = null;
    let existing_key_expiration = null;
    // Check if there is at least one existing key associated with the pass codes
    for (let i = 0; i < response.length; i++) {
      if (response[i]) {
        existing_key = response[i];
        existing_key_expiration = await redis_client.expiretime(key_cache_redis_keys[i]);
        break;
      }
    }
    if (existing_key) {
      // If there is we will return it and store this key to all the empty keys
      for (let i = 0; i < response.length; i++) {
        if (!response[i]) {
          await redis_client.set(key_cache_redis_keys[i], existing_key);
          await redis_client.expireat(key_cache_redis_keys[i], existing_key_expiration);
        }
      }
      res.json({
        status: "SUCCESS",
        metager_pass_key: {
          key: existing_key,
          valid_until: dayjs.unix(existing_key_expiration).format()
        }
      });
    } else {
      // There is no Key yet. Let's generate one
      let expiration = (new dayjs()).add(6, 'hour').unix();

      Key.CREATE_NEW_KEY(expiration).then(async key => {
        for (let i = 0; i < key_cache_redis_keys.length; i++) {
          await redis_client.set(key_cache_redis_keys[i], key);
          await redis_client.expireat(key_cache_redis_keys[i], expiration);
        }
        res.json({
          status: "SUCCESS",
          metager_pass_key: {
            key: key,
            valid_until: dayjs.unix(expiration).format()
          }
        });
      }).catch(reason => {
        res.status(400).json({
          status: "FAILURE",
          msg: reason
        });
      });
    }
  }).catch(reason => {
    res.status(400).json({
      status: "FAILURE",
      msg: reason
    });
  });


});

module.exports = router;
